import sys, os
import cStringIO
verbose = False


class Atom :
	'''
	Atom class contains common information about a line in PDB coordinate section.
	'''
	def __init__( self ) :
		self.residue = None

		self.atom_line = None
		self.type = None

		self.number = 0
		self.name = ""
		self.insertion_code = None
		self.alternative_position = None
		self.coordinate = None
		self.bfactor = None
		self.occupancy = None
		self.element = ""
		self.charge = ""

	def get_residue_id( self ) :
		try: 
			n = int(self.atom_line[22:26])
			resnum = self.atom_line[22:26]
			chain_id = self.atom_line[21]
			insertion_code = self.atom_line[26]
			return resnum +'\t'+ chain_id +'\t'+ insertion_code
		except :
			return None

	def get_type( self ) :
		return self.type

	def get_number( self ) :
		return self.number

	def parse( self, fp=None, atom_line=None ) :
		if fp :
			atom_line = fp.readline()
		elif atom_line :
			pass
		else :
			return

		self.atom_line=atom_line
		self.type = atom_line[:6]
		
		try :
			self.number = int(atom_line[6:11])
		except :
			return

		return 1
	
	def get_chain_id( self ) :
		return self.atom_line[21]

	def get_residue_name( self ) :
		return self.atom_line[17:20]

	def is_ca_atom( self ) :
		if self.atom_line[12:16] == ' CA ' :
			return True
		else :
			return False 


class Residue :
	def __init__( self ) :
		self.atoms = []
		self.id = None
		self.name = None
		self.number = 0

	def get_name( self ) :
		return self.name

	def get_name_letter( self ) :
		three2one_common = { 'GLY':'G', 'PRO':'P', 'ALA':'A', 'VAL':'V', 'LEU':'L', 'ILE':'I', 'MET':'M', 'MSE':'M', 'CYS':'C', 'PHE':'F', 'TYR':'Y', 'TRP':'W', 'HIS':'H', 'LYS':'K', 'ARG':'R', 'GLN':'Q', 'ASN':'N', 'GLU':'E', 'ASP':'D', 'SER':'S', 'THR':'T' }
		if self.name in three2one_common :
			return three2one_common[self.name]
		else :
			return 'X'

	def is_part_of_protein( self ) :
		for atom in self.atoms :
			if atom.is_ca_atom() :
				return True
		return False

	def add_atom( self, atom ) :
		self.atoms.append( atom )

		if not self.name :
			self.name = atom.get_residue_name()
		if not self.id :
			self.id = atom.get_residue_id()
			try :
				self.number = int(self.id[:4])
			except :
				print '"'+self.id[:4]+'"', self.id
				sys.exit()

	def get_number( self ) :
			return self.number


class Chain :
	'''
	Chain class contains list of residues that are in the same chain.
	'''
	def __init__( self, model=None )  :
		self.model = model #pointer to the model that contains this chain
		self.id = None #chain id

		self.residue_ids = []
		self.residues = {}
		self.residues_within_chain = set()
		
		self.terminated = False #variable finding the info for the chain
		
	def get_residue( self, residue_id ) :
		if residue_id in self.residues :
			return self.residues[residue_id]

	def get_within_chain_residue_ids( self ) :
		residue_ids = []
		for id in self.residue_ids :
			if self.is_residue_within_chain( id ) :
				residue_ids.append( id )

		return residue_ids

	def is_residue_within_chain( self, residue_id ) :
		if  residue_id in self.residues_within_chain :
			return True
		else :
			return False

	def get_residue_within_chain( self, residue_id ) :
		if residue_id in self.residues_within_chain :
			return self.residues[ residue_id ]

	def add_residue( self, residue_id, atom_line ) :
		residue = None
		if residue_id in self.residues :
			residue = self.residues[residue_id]
		else :
			residue = Residue()
			residue.residue_id = residue_id
		
			self.residues[residue_id] = residue
			self.residue_ids.append( residue_id )

		if atom_line[:6] == 'TER   ' :
			self.terminated = True
		
		if self.terminated == False :
			self.residues_within_chain.add(residue_id)

		return residue
	


class Model :
	def __init__( self, pdb=None ) :
		self.pdb = pdb

		self.number = None
		self.chain_ids = [] #keeps appearing order
		self.chains = {}

		self.atoms = []

	def get_chain( self, chain_id ) :
		if chain_id in self.chains :
			return self.chains[chain_id]

	def add_chain( self, chain_id ) :
		chain = None
		if chain_id in self.chains :
			chain = self.chains[ chain_id ]
		else :
			chain = Chain()
			self.chains[chain_id] = chain
			self.chain_ids.append( chain_id )

		return chain

	def add_atom( self, atom ) :

		self.atoms.append( atom )
		residue_id = atom.get_residue_id()

		if residue_id :
			residue = None

			chain_id = atom.get_chain_id()
			chain = self.add_chain( chain_id )
			residue = chain.add_residue( residue_id, atom.atom_line )
			residue.add_atom( atom )
			
			



	def parse( self, fp ) :
	 	for line in fp :
			atom = Atom()
			if not atom.parse( atom_line=line ) :
				continue

			self.add_atom( atom )


class PDB :
	def __init__( self, pdb_fn ) :
		self.model_ids = [] #keeps model order
		self.models = {}
		
		self.pdb_fn = pdb_fn
		if os.path.exists( pdb_fn ) :
			self.parse()

		self.residue_selections = []

	
	def get_model( self, model_id ) :
		'''
		Get model correponding the given model_id
		if the model_id is not in the model_id list,
		the first model will be returned
		'''
		if model_id in self.models :
			return self.models[model_id]
		else :
			return self.models[ self.model_ids[0] ]


	def select_residues_by_scop_range( self, scop_range ) :
		'''
		This function select range of scop
		'''
		ranges = scop_range.split(',')
		for r in ranges :
			chain_id = None
			start_resnum = None
			end_resnum = None

			chain_and_numbers = r.split(':')
			chain_id = chain_and_numbers[0]
			if chain_and_numbers[1] :
				number_array = chain_and_numbers[1].split('-')
				numbers = []
				while number_array :
					i = number_array.pop(0)
					if i :
						numbers.append(int(i))
					else :
						numbers.append( -1*int(number_array.pop(0)) )

				start_resnum, end_resnum = numbers
					
			if not self.select_residues( chain_id=chain_id, start_resnum=start_resnum, end_resnum=end_resnum) :
				return False 

		return True
	
			
	def check_selected_residue( self, residue ) :
		for selection in self.residue_selections :
			if residue in selection :
				return True
		return False

	def select_residues( self, model_id=0, chain_id='A', start_resnum=None, end_resnum=None ) :
		selection = []
		chain = self.get_model( model_id ).get_chain( chain_id )

		if not chain :
			print >>sys.stderr, "Warning: No Chain", chain_id, "exists."
			return False

		for residue_id in chain.get_within_chain_residue_ids() :
			residue = chain.get_residue( residue_id )
			resnum = residue.get_number()
			if start_resnum == None or resnum >= start_resnum :
				if end_resnum == None or resnum <= end_resnum :
					if not self.check_selected_residue( residue ) :
						selection.append( residue )
					else :
						print >>sys.stderr, 'WARNING: Duplicated Selection', residue_id
						return False

		if selection :
			self.residue_selections.append( selection )

		return True

	def reset_selection( self ) :
		self.residue_selections = []

	def selection2sequence( self ) :
		str_fp = cStringIO.StringIO()
		if verbose :
			print self.residue_selections
		for selection in self.residue_selections :
			for residue in selection :
				#print residue.id, residue.get_name(), residue.get_name_letter()
				str_fp.write( residue.get_name_letter() )
		return str_fp.getvalue()
	

	def selection2fragment_range_string( self ) :
		fragment_lengths = []
		if verbose :
			print "@ selection2fragment_range_string"
			
		for selection in self.residue_selections :
			str_fp = cStringIO.StringIO()
			for residue in selection :
				str_fp.write( residue.get_name_letter() )
			fragment_lengths.append( len(str_fp.getvalue()) )

		if not self.residue_selections :
			print >>sys.stderr, "Error! No residues selected!"
			return ""

		if verbose :
			print 'fragment_lengths:',fragment_lengths

		line = '0-' + str(fragment_lengths[0])
		current_offset = fragment_lengths[0]
		for length in fragment_lengths[1:] :
			line += ' ' + str(current_offset) + '-'+str(length+current_offset)
			current_offset += length

		return line
			

	def parse( self ) :
		fp = open( self.pdb_fn )
		model_fp = None
		model_number = 0
		for line in fp :
			if line[:6] == "MODEL " :
				model_fp = cStringIO.StringIO()
				model_number = int(line[10:14])
				continue
			
			if line[:6] == 'ENDMDL' :
				if not model_fp :
					print >>sys.stderr, 'Error while reading PDB Models'
					return
	
				#adding the Model
				model_fp.flush()
				model_fp.seek(0) #rewind the string

				model = Model( pdb=self )
				model.number = model_number

				model.parse( model_fp )
				if not model_number in self.models :
					self.models[model_number] = model
					self.model_ids.append( model_number )
				else :
					print >>sys.stderr, 'WARNING model_number has been duplicated!', model_number, self.model_ids

				model_fp.close()
				model_fp = None
				continue

			if model_fp :
				model_fp.write( line )
			#for cases where the PDB file only contains single model without MODEL definition
			elif line[:6] == 'ATOM  ' or line[:6] == 'HETATM' :
				model_fp = cStringIO.StringIO()
				model_fp.write(line)

		else :
			if model_fp :
				model_fp.flush()
				model_fp.seek(0) #rewind the string

				model = Model( pdb=self )
				model.number = model_number

				model.parse( model_fp )
				if not model_number in self.models :
					self.models[model_number] = model
					self.model_ids.append( model_number )
				else :
					print >>sys.stderr, 'WARNING model_number has been duplicated!', model_number, self.model_ids


				model_fp = None
			
		fp.close()


def generate_inserted_scop_ranges( scop_range ) :
	'''
	This function takes scop_range and returns list of scop_ranges for inserted domains
	'''
	inserted_scop_ranges = []
	#need to return
	ranges = scop_range.split(',')
	if len(ranges ) < 2 :
		print >>sys.stderr, "Error! No inserted domains.", scop_range
		return inserted_scop_ranges

	for r1, r2 in zip(ranges[:-1], ranges[1:]) :
		chain_id1 = None
		start_resnum1 = None
		end_resnum1 = None

		chain_id2 = None
		start_resnum2 = None
		end_resnum2 = None

		chain_and_numbers1 = r1.split(':')
		chain_id1 = chain_and_numbers1[0]
		if chain_and_numbers1[1] :
			number_array = chain_and_numbers1[1].split('-')
			numbers = []
			while number_array :
				i = number_array.pop(0)
				if i :
					numbers.append(int(i))
				else :
					numbers.append( -1*int(number_array.pop(0)) )

			start_resnum1, end_resnum1 = numbers

		chain_and_numbers2 = r2.split(':')
		chain_id2 = chain_and_numbers2[0]
		if chain_and_numbers2[1] :
			number_array = chain_and_numbers2[1].split('-')
			numbers = []
			while number_array :
				i = number_array.pop(0)
				if i :
					numbers.append(int(i))
				else :
					numbers.append( -1*int(number_array.pop(0)) )

			start_resnum2, end_resnum2 = numbers

		if chain_id1 == chain_id2 :
			pass
		else :
			print >>sys.stderr, 'No inserted domain found, the domain should come from the same chain!', scop_range
			continue
				
		inserted_scop_ranges.append( '%s:%s-%s'%( chain_id1, end_resnum1+1, start_resnum2-1 ) )
		

	return inserted_scop_ranges


		

if __name__ == '__main__' :
	pdb_fn = sys.argv[1]
	pdb = PDB( pdb_fn )
	pdb.select_residues( chain_id='A', start_resnum=801, end_resnum=810 )
	print pdb.selection2sequence()
	print pdb.selection2fragment_range_string()

	pdb.reset_selection()
	pdb.select_residues_by_scop_range( 'A:,B:' )
	print pdb.selection2sequence()
	print pdb.selection2fragment_range_string()

